from UNAGI import UNAGI
unagi = UNAGI()
unagi.analyse_UNAGI('/mnt/md0/yumin/to_upload/UNAGI/data/small/2/stagedata/org_dataset.h5ad',2,10,target_dir=None,customized_drug='/mnt/md0/yumin/to_upload/UNAGI/data/jasper_target_pair.npy')